import random
import re
import os
import numpy as np
from typing import Tuple
from addict import Dict
import string
import requests
import ast
from matplotlib import pyplot as plt
import csv


class Generator:

    @staticmethod
    def unique_matchId_generator(letter_count: int = 12, digit_count: int = 8) -> str:
        str1 = ''.join((random.choice(string.ascii_letters) for x in range(letter_count)))
        str1 += ''.join((random.choice(string.digits) for x in range(digit_count)))

        sam_list = list(str1)  # it converts the string to list.
        random.shuffle(sam_list)  # It uses a random.shuffle() function to shuffle the string.
        final_string = ''.join(sam_list)
        return final_string

    @staticmethod
    def path_url_creator(**kwargs) -> Tuple[str, str]:
        """

        :param params.gui:
        :param params.cfg:
        :param params.splitData:
        :param params.index:
        :param params.directory:
        :return:
        """
        params = Dict(kwargs)
        i = params.index
        host = params.cfgImage.ip_remote if params.cfgImage.Remote else params.cfgImage.ip_local
        if params.gui:
            ip_path = re.compile("^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
            ip_check = bool(ip_path.match(params.directory))
            if ip_check:
                path = os.path.join(
                    "http://" + host + ":" + params.cfgImage.port + "/" + params.cfgImage.Directory + ":",
                    str(params.splitData[i]["dirname"]), params.splitData[i]["filename"], params.splitData[i]["imgname"])
            else:
                path = os.path.join(params.directory, str(params.splitData[i]["dirname"]),
                                    params.splitData[i]["filename"],
                                    params.splitData[i]["imgname"])
        else:
            # TODO will be dynamic connection for now test phase
            if params.cfgImage.device == "google":
                #     H:\IMAGERY\mapilio_dev\organization\arnavutkoy
                #     https://image.mapilio.com/h:/IMAGERY/mapilio_dev/organization/arnavutkoy/TqbHYpgLXzU7vHKWQDRVjQ.jpeg
                path = os.path.join("https://" + "cdn.mapilio.com", "h:", "IMAGERY", 'mapilio_dev', 'organization',
                                    'arnavutkoy', params.splitData[i]["imgname"])
            if params.cfgImage.device == 'ladybug':
                path = os.path.join(
                    "http://" + "cdn.mapilio.com/" + str(params.cfgImage.directory) + ":",
                    str(params.splitData[i]["dirname"]), params.splitData[i]["filename"], params.splitData[i]["imgname"])

        return host, path

    @staticmethod
    def data_separation(data: list, dividing_percentage: int) -> list:
        """
         Segmenting incoming data

        :param data: data to be processed
        :param dividing_percentage: percentage of data fragmented
        :return: predicted masks
        """
        percentage = int(len(data) / 100 * dividing_percentage)

        for i in range(0, len(data), percentage):
            # Create an index range for l of n items:
            yield data[i:i + percentage]

    @staticmethod
    def take_objects(matchedObjects: list, take: int = 2):

        for i in range(0, len(matchedObjects), take):
            if len(matchedObjects[i:i + take]) % take == 0:
                yield matchedObjects[i:i + take]

    @staticmethod
    def get_exif_information(img_info):
        """
        :param img_info: exif object
        :return: (lat, lon), orientation, (Height, Width), FocalLength, Altitude,
        """
        information = {}
        data = img_info.extract_exif()
        try:
            information["model"] = data["model"]
            information["coordx"] = data["gps"]["latitude"]
            information["coordy"] = data["gps"]["longitude"]
            information["width"] = data["width"]
            information["height"] = data["height"]

            # Focal Length
            fLen_obj = data["gps"]["FocalLength"]
            fLen_str = f"{fLen_obj}"
            fLen_arr = fLen_str.split("/")
            fLen = float(int(fLen_arr[0]) / int(fLen_arr[1]))
            information["FocalLength"] = fLen

            hor_width = data["height"] if data["orientation"] == 1 else data["width"]
            information["orientation"] = hor_width
            # Angle of View
            aFov = np.arctan(hor_width / (2 * fLen)) * (180 / np.pi)
            information["FoV"] = aFov
        except:
            raise Exception(f"Check the image Exif Data some missing values")

        for k, v in information.items():
            if information[k] is None:
                raise Exception(f"{k} is None")
            else:
                pass

        return information

    @staticmethod
    def get_config_response(configUrl: str) -> Dict:
        response = str(Dict(requests.get(configUrl).json()).config)
        pure_response = response.replace("\'", "\"")
        fix_dict = Dict(ast.literal_eval(pure_response))
        return fix_dict

    @staticmethod
    def get_random_hex_color():
        import random
        random_number = random.randint(0, 16777215)
        hex_number = str(hex(random_number))
        color = '#' + hex_number[2:]
        return color

    @staticmethod
    def plt_create_detected_unique_points(
            file_id,
            **points,
            ):
        """
        Creating detected unquie points figure
        Args:
            points : detected_unique_points: classnames and count
                    detected_single_points: single classnames and count
            file_id: where will be saving

        Returns:

        """

        headers = ['Name', 'Count']

        for key, value in points.items():
            for name, detected_points in value.items():
                unique, counts = np.unique(detected_points, return_counts=True)
                detected_points_counts_dict = dict(zip(unique, counts))

                with open(os.path.join('Exports', file_id, f'{name}.csv'), 'w') as csvfile:
                    writer = csv.writer(csvfile)
                    writer.writerow(headers)
                    for key_, value_ in detected_points_counts_dict.items():
                        writer.writerow([key_, value_])

                y_pos = np.arange(len(unique))

                plt.figure(figsize=(44, 22))
                plt.barh(y_pos, counts, align='center', alpha=0.5)
                plt.yticks(y_pos, unique)
                plt.xlabel('Counts')
                plt.title(f"{name}")
                plt.savefig(os.path.join('Exports', file_id, f'{name}.png'))  # noqa
                plt.clf()



import json
from typing import Tuple

import numpy as np
import pandas as pd
import requests
from trianglesolver import solve, degree
from addict import Dict
import os
import csv
from urllib.parse import urlparse
# from imagery.pixel import Pixel
import logging
logger = logging.getLogger(__name__)


class Utilities:

    # TODO: Will be moved to Pixel Lib
    # @staticmethod
    # def check_object_feature(bbox: list, feature: dict) -> dict:
    #     """
    #
    #     Args:
    #         bbox: detected object bbox such as shop sign, utility pole
    #         feature: probability feature such as plate, mark
    #
    #     Returns:
    #         If there is a desired feature in the detected object, it writes it to the dict.
    #         Examples: { 'plate' : True }
    #     """
    #
    #     if feature is not {}:
    #         for key, valBbox in feature.items():
    #             for class_name, bboxB in valBbox.items():
    #                 if Pixel.calc_bb_intersection_over_union(boxA=bbox, boxB=bboxB):
    #                     return {class_name: True}
    #                 else:
    #                     return {}
    #     else:
    #         return {}

    @staticmethod
    def is_match(matched_objects: list) -> bool:
        """
        matched_objects matched objects twice
        :param matched_objects:
        :return: check out length bool type
        """
        if len(matched_objects) > 0:
            return True
        else:
            return False

    @staticmethod
    def is_between(t1max: bool, t2max: bool) -> bool:
        """
        the aim check between each detected thetas

        :param t1max: t1max is thetas max 1  between defined angle wide in config
        :param t2max: t2max is thetas max 1  between defined angle wide in config
        :return: check cross match as bool type
        """
        return True if (t1max and t2max) else False

    @staticmethod
    def is_right_or_left(**kwargs) -> bool:
        """
        Goal : detected objects must be same region each right or left.
        # theta1: float, heading1: float, theta2: float, heading2: float
        :param theta1: first object which seen by car angle
        :param heading1: moment car location angle according to north
        :param theta2: second object which seen by car angle
        :param heading2: moment car location angle according to north
        :return: gives answer as bool type objects left or right check
        """
        params = Dict(**kwargs)
        theta1 = params.theta1
        theta2 = params.theta2
        heading1 = params.heading1
        heading2 = params.heading2

        if theta1 <= heading1 and theta2 <= heading2:
            return True
        elif theta1 >= heading1 and theta2 >= heading2:
            return True
        else:
            return False

    @staticmethod
    def check_validity(a, b, c):
        """
         Triangle rule checks
        it's triangle edged as a, b, c
         5 < C / degree < 20 must be otherwise return false

        :param a: detected point and **first** car position edge
        :param b: detected point and **second** car position edge
        :param c: first and second car edge
        :return: if is not triangle and C degree over the defined variable return False.
        """

        a, b, c, A, B, C = solve(a, b, c)
        if (a + b <= c) or (a + c <= b) or (b + c <= a):
            if not 5 < (C / degree) < 20:
                return False
        return True

    @staticmethod
    def get_label_name_or_id(labelId: int, config: dict):
        """
         this function has two assignment

        :param labelId: if labelId is **array** return classname else
                        labelId is **string** return classname id
        :param config:
        :return: object's name or id
        """

        with open(config.labeljson, 'r') as f:
            labelsDict = json.load(f)
            if isinstance(labelId, np.ndarray):
                return labelsDict.get(str(labelId))
            elif isinstance(labelId, str):
                for id, name in labelsDict.items():
                    if name == labelId:
                        return id

    @staticmethod
    def file_exist_check_and_open(path: str):
        if os.path.exists(path):
            append_write = 'a'  # append if already exists
        else:
            append_write = 'w'  # make a new file if not

        return open(path, append_write)

    @staticmethod
    def file_check(file_id: str):
        if not os.path.exists(os.path.dirname("Exports")):
            try:
                os.makedirs("Exports")
            except OSError as err:
                logger.info(err)
        os.makedirs(os.path.join("Exports", file_id, "ObjectsImage"))
        os.makedirs(os.path.join("Exports", file_id, "imgs"))

        header = ['ImagePath', 'ImageName']
        with open(os.path.join("Exports", file_id, "images_list.csv"), 'w', encoding='UTF8') as f:
            writer = csv.writer(f)
            writer.writerow(header)
            f.close()
        outPath = os.path.join("Exports", file_id)

        return outPath

    @staticmethod
    def config_reader(path: str) -> Dict:

        with open(path) as f:
            cfg = json.load(f)

        return Dict(cfg)

    @staticmethod
    def image_logger(url: str, file_id: str) -> bool:
        """

        Args:
            url: remote image url
            file_id: current processing task

        Returns: have not taken image false, otherwise true

        """
        df = pd.read_csv(os.path.join('Exports', file_id, 'images_list.csv'))

        url_check = df[df["ImagePath"] == url]
        if url_check.empty:
            url_parse = urlparse(url)
            image_name = os.path.basename(url_parse.path)
            df2 = {'ImagePath': url, 'ImageName': image_name}
            df = df.append(df2, ignore_index = True)
            df.to_csv(os.path.join('Exports', file_id, 'images_list.csv'), index=False)
            return False
        else:
            return True

    @staticmethod
    def predict_exclusive_inclusive(
            use_predict: list,
            not_use_predict: list,
            predicted_class: str
    ) -> Tuple[bool, list]:
        """

        Args:
            use_predict: will be detecting objects classes
            not_use_predict: not use for detecting classes
            predicted_class: model predicted class
        Returns:

        """

        for g in use_predict:
            if g in not_use_predict:
                not_use_predict.remove(g)

        result = any(elem == predicted_class or elem == 'all' for elem in use_predict)

        if result:
            return True, not_use_predict
        else:
            return False, not_use_predict

    @staticmethod
    def is_url_image(url: str):
        r = requests.get(url, stream=True)
        if r.status_code == 200:
            return True
        else:
            return False
